//---------------------------------------------------------------------------
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#define minOfThree(x1,x2,x3) x1 < x2 ? x1 < x3 ? x1 : x3 : x2 < x3 ? x2 : x3
#define max(A,B) A > B ? A : B
#define min(A,B) A < B ? A : B
//---------------------------------------------------------------------------

unsigned n;
int w=32;

typedef struct Partition {
   int a;
   int b;
   int error;
} PartStruct;

typedef struct SAElement {
   unsigned pos;
   unsigned char depth;
   unsigned char jump;
} TSAElement;

typedef struct MSAJump {
   unsigned Mv;
   unsigned Pv;
   int Score;
} TMSAJump;

typedef struct MCactusS {
   unsigned Mv;
   unsigned Pv;
   int Score;
} TMCactusState;

typedef struct MCactusSU {
   unsigned *Mv;
   unsigned *Pv;
   unsigned *Score;
   int block;
} TMCactusStateUnstricted;

typedef struct NCactusS {
   unsigned D;
   unsigned zerodiag;
} TNCactusState;

typedef struct DPCactusS {
   int *C;
   int limit;
} TDPCactusState;

class TSTreeNode
{
public:
   unsigned minindex; // ensimminen merkki, jonka solmu sislt
   TSTreeNode* branchBroather; // linkkilista solmun veljist
   unsigned GetMinindex() {return (minindex << 1) >> 1;};
   unsigned SetMinindex(unsigned m) {minindex = m | ((minindex >> ((w-1))) << (w-1));};
   bool Small() {return !(minindex >> (w-1));};
   TSTreeNode() { minindex = 0; };
};


class TSTreeInNode : TSTreeNode
{
public:
   TSTreeNode* suffixLink; // linkki suffiksiin
   TSTreeNode* firstChild; // linkki ensimmiseen lapseen
   TSTreeInNode() { TSTreeNode(); minindex = 1<<(w-1);
                    suffixLink = NULL; branchBroather = NULL;
                    firstChild = NULL;
                  };
   unsigned Maxindex() {if (firstChild == NULL) return n; else return firstChild->GetMinindex()-1;};
   ~TSTreeInNode() {if (firstChild != NULL) delete firstChild;
                    if (branchBroather != NULL) delete branchBroather; };
};

class TSTreeInSmallNode : TSTreeNode
{
public:
   TSTreeNode* firstChild; // linkki ensimmiseen lapseen
   TSTreeInSmallNode() { TSTreeNode(); minindex = 1 << (w-1);
                    branchBroather = NULL;
                    firstChild = NULL;
                  };
   unsigned Maxindex() {if (firstChild == NULL) return n; else return firstChild->GetMinindex()-1;};
   ~TSTreeInSmallNode() {if (firstChild != NULL) delete firstChild;
                    if (branchBroather != NULL) delete branchBroather; };
};

class TSTreeLeaf : TSTreeNode
{
public:
   unsigned Maxindex() {return n;};
   TSTreeLeaf() {TSTreeNode();};
};

class TMatchList {
   unsigned Size;
public:
   unsigned *Data;
   unsigned Top;
   TMatchList() {Top = 0; Size = 100; Data = (unsigned *)malloc(Size*sizeof(unsigned));};
   ~TMatchList() {free(Data);};
   void Clear() {Top = 0;};
   int IndexOf(unsigned pos)
   {
      int i;
      for(i=0; i< Top; i++)
         if (Data[i] == pos)
            return i;
      return -1;
   };
   void Add(unsigned pos)
   {
     if (Top >= 100000)
        return;
     if (Size <= Top+1) {
        Size *=2;
        Data = (unsigned *)realloc(Data,Size*sizeof(unsigned));
     }
     Data[Top++] = pos;
   };
   unsigned& operator[](int i) { return Data[i];};
   //const int& operator[](int i) const { return Data[i];};
};

class TUnsignedObject {
public:
   TUnsignedObject *next;
   unsigned x;
};

class TCactusStack {
private:
   TUnsignedObject *first;
public:
   bool Empty() {return first == NULL;};
   void Push(unsigned x)
           {
            TUnsignedObject *s= new TUnsignedObject();
            s->next = first;
            first = s;
            s->x = x;
           };
   unsigned Pop()
           {
            unsigned Result;
            TUnsignedObject *temp = first;
            Result = temp->x;
            first = temp->next;
            delete temp;
            return Result;
           };
   TCactusStack() {first = NULL;};
};


TMatchList *MatchList = new TMatchList();
TMatchList *MatchingOnes;
TSTreeNode* root;
TSAElement *SuffixArray;
char* text;
char* pattern;
unsigned nodeCount = 0;
unsigned int **Peq = NULL;
unsigned int B[256];
bool S[256];
int V[265];
unsigned Nodes[1000];
unsigned k = 1;
unsigned m;
unsigned int Din,D,M1,M2,M3,G;
unsigned int zerolast;
unsigned int zerofirst;
PartStruct ** PP=NULL;
int part_r=0;
unsigned int Eq, Xv, Xh, Ph, Mh;
unsigned int twopowmminusone;
unsigned int blockCount;
unsigned int block;
unsigned int lastBlockLength;
unsigned int *Pv;
unsigned int *Mv;
unsigned int *Score;
unsigned int twopowmpos;
unsigned int twopowwpos;
bool print=false;
unsigned nodesVisited=0;
unsigned sumdepths=0;
unsigned sumstops=0;
unsigned sumnodedepths=0;
bool debug=true;

unsigned *SuffixCactus;
unsigned *SiblingCactus;
unsigned char *DepthCactus;


void MPreComp(char *P);
void MPreCompUnstricted(char *P);
int MComp();
int MCompFilter();
int MCompUnstricted();
void MCompDistance();
void NPrecomp(char *P);
int NComp();
void NCompDist();
void MakeSuffixTree();
int MSearchSuffixTree();
int NSearchSuffixTree();
int MSearchSuffixCactus();
int NSearchSuffixCactus();
void PrintTree(TSTreeNode *s, int depth);
int DPSearch(char *P);
void PatternPartition(int rr);
int MPartSearchST(char *P);
int NPartSearchST(char *P);
int MPartSearchSC(char *P);
int NPartSearchSC(char *P);
int DPSearchST(char *P);
int DPPartSearchST(char *P);
int DPSearchSC(char *P);
int DPPartSearchSC(char *P);
void MakeSuffixCactus();
void PrintSuffixArray();
int MSearchSuffixArray();
int CalcOptPart(unsigned nn, unsigned mm, unsigned kk, char *buf);
char * SubString(char * A,unsigned start, unsigned count);
void ShowMatchesFromCactus(unsigned j, unsigned depth);


void makeindex(char *buf,unsigned nbuf,int alg,int qval)
{
   print = qval;
   text = buf;
   n = nbuf;
   if (alg >= 10 && alg < 20)
      MakeSuffixTree();
   else if (alg >=20 && alg < 30)
      MakeSuffixCactus();
}

void prep(char *s, unsigned mval, int alg, int pp_r, int kval, int qval)
{
   m = mval;
   k = kval;
   print = qval;
   pattern = SubString(s,1,m);
   if (alg % 10 == 1 || alg % 10 == 2)
      MPreComp(s);
   else if (alg % 10 == 3)
      MPreCompUnstricted(s);
   else if (alg % 10 == 4)
      NPrecomp(s);
   else if (alg % 10 == 5 || alg % 10 == 6)
      if (pp_r > 1)
         PatternPartition(pp_r);
      else
         if (alg % 10 == 6)
            NPrecomp(s);
         else if (alg % 10 == 5 && m <= w)
            MPreComp(s);
         else
            MPreCompUnstricted(s);
}

int exec(char *buf, unsigned len, int alg, int pp_r, int kval, int qval)
{
   k = kval;
   print = qval;
   text = buf;
   n = len;
   if (alg == 1)
      return MComp();
   else if (alg == 2)
      return MCompFilter();
   else if (alg == 3)
      return MCompUnstricted();
   else if (alg == 4)
      return NComp();
   else if (alg == 7)
      return DPSearch(pattern);
   else if ((alg >= 11 && alg <= 13)  || (pp_r == 1 && alg == 15))
      return MSearchSuffixTree();
   else if (alg == 14 || (pp_r == 1 && alg == 16))
      return NSearchSuffixTree();
   else if (alg == 15)
      return MPartSearchST(pattern);
   else if (alg == 16)
      return NPartSearchST(pattern);
   else if (alg == 17)
      return DPSearchST(pattern);
   else if (alg == 18)
      return DPPartSearchST(pattern);
   else if (alg >= 21 && alg <=23)
      return MSearchSuffixCactus();
   else if (alg == 24)
      return NSearchSuffixCactus();
   else if (alg == 25)
      return MPartSearchSC(pattern);
   else if (alg == 26)
      return NPartSearchSC(pattern);
   else if (alg == 27)
      return DPSearchSC(pattern);
   else if (alg == 28)
      return DPPartSearchSC(pattern);
}

// Myers
void MPreComp(char *P) {
   int i;
   int j;
   Peq = (unsigned **)malloc((((m-1) / w)+1)*sizeof(unsigned*));
   int l;
   for (l=0;l <= ((m-1) / w); l++)
      Peq[l] = (unsigned *)malloc(256*sizeof(unsigned));
   for(i = 0;i< 256;i++) {
      S[i] = false;
      for (l=0;l <= ((m-1) / w); l++)
         Peq[l][i] = 0;
   }
   for (l=0;l <= ((m-1) / w); l++)
      for (j = 1; j <= w; j++) {
         if (w*l+j > m)
            break;
         Peq[l][P[w*l+j-1]] = Peq[l][P[w*l+j-1]] | (1 << (j-1));
         if (j+w*l <= k+1)
            S[P[w*l+j-1]] = true;
      }
}

void MPreCompUnstricted(char *P)
{
   MPreComp(P);
   blockCount= ((m-1) / w);
   block = (k-1) / w;
   lastBlockLength = m % w;
   if (lastBlockLength == 0)
      lastBlockLength = w;
   Pv = (unsigned *)malloc((blockCount+1)*sizeof(unsigned));
   Mv = (unsigned *)malloc((blockCount+1)*sizeof(unsigned));
   Score = (unsigned *)malloc((blockCount+1)*sizeof(unsigned));
   twopowmpos = (unsigned)1 << (lastBlockLength-1);
   twopowwpos = (unsigned)1 << (w-1);
   int l;
   for (l=0; l < blockCount; l++) {
      Mv[l] = 0;
      Score[l] = (l+1)*w;
      Pv[l] = (~(unsigned)0);
   }
   Mv[blockCount] = 0;
   Score[blockCount] = m;
   Pv[blockCount] = (~(unsigned)0) >> ((blockCount+1)*w - m);
}

int MComp()
{
   unsigned int Mv = 0;
   unsigned int Pv = (~(unsigned)0) >> (w - m);
   unsigned int Score = m;
   twopowmminusone = (unsigned)1 << (m-1);
   int j;
   int count=0;
   for(j = 1; j <= n-1; j++) {
      Eq = Peq[0][(unsigned char)text[j-1]];
      Xv = Eq | Mv;
      Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
      Ph = Mv | ~ (Xh | Pv);
      Mh = Pv & Xh;
      if (Ph & twopowmminusone)
         Score += 1;
      else if (Mh & twopowmminusone)
         Score -= 1;
      Ph <<= 1;
      Pv = (Mh << 1) | ~ (Xv | Ph);
      Mv = Ph & Xv;
      if (Score <= k) {
         count++;
         if (print)
            printf("Match at %d\n",j);
      }
   }
   return count;
}

int MCompFilter()
{
   unsigned int Mv = 0;
   unsigned int Pv = (~(unsigned)0) >> (w - m);
   unsigned int Score = m;
   twopowmminusone = (unsigned)1 << (m-1);
   int j;
   int count=0;
   for(j = 1; j <= n-1; j++) {
    if (Score != m || S[text[j-1]]) {
      Eq = Peq[0][(unsigned char)text[j-1]];
      Xv = Eq | Mv;
      Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
      Ph = Mv | ~ (Xh | Pv);
      Mh = Pv & Xh;
      if (Ph & twopowmminusone)
         Score += 1;
      else if (Mh & twopowmminusone)
         Score -= 1;
      Ph <<= 1;
      Pv = (Mh << 1) | ~ (Xv | Ph);
      Mv = Ph & Xv;
      if (Score <= k) {
         count++;
         if (print)
            printf("Match at %d\n",j);
      }
    }
   }
   return count;
}

int MCompUnstricted()
{
   int count=0;
   int blockAverage=0;
   unsigned Phendbit;
   unsigned Mhendbit;
   unsigned temp;
   int j;
   int l;
   unsigned overk=1;
   block = (k-1) / w;
   for(j = 1; j <= n; j++) {
    Phendbit = 0;
    Mhendbit = 0;
    for (l=0; l <= blockCount; l++) {
      Eq = Peq[l][(unsigned char)text[j-1]];
      Xv = Eq | Mv[l];
      Xh = ((((Eq | Mhendbit)& Pv[l]) + Pv[l]) ^ Pv[l]) | (Eq | Mhendbit);
      Ph = Mv[l] | ~ (Xh | Pv[l]);
      Mh = Pv[l] & Xh;
      temp = l < blockCount ? twopowwpos : twopowmpos;
      if (Ph & temp)
         Score[l] += 1;
      else if (Mh & temp)
         Score[l] -= 1;
      temp = (Ph >> (w-1));
      Ph <<= 1;
      Ph |= Phendbit;
      Phendbit = temp;
      temp = (Mh >> (w-1));
      Mh <<= 1;
      Mh |= Mhendbit;
      Mhendbit = temp;
      Pv[l] = (Mh) | ~ (Xv | Ph);
      Mv[l] = Ph & Xv;
      if (block == l+1 &&
          ((block == blockCount &&
            (Score[l] > k+lastBlockLength ||
             (Score[l] > k && overk != 0 && j - overk >= lastBlockLength))) ||
           (block != blockCount &&
            (Score[l] > k+w ||
             (Score[l] > k && overk != 0 && j - overk >= w)))))  {
         // lohkon Score kasvanut niin isoksi, ettei seuraavaa kannata laskea
         overk = 0;
         block = l;
         break;
      }
      else if (block == l+1 && overk == 0 && Score[l] > k)
         // Talletetaan ensimminen diagonaali jolla Score > k. Nin tiedetn
         // jatkossa koska seuraavan lohkon laskeminen on turhaa.
         overk = j;
      else if (block == l+1 && Score[l] <= k)
         // Diagonaalilla Score <= k ==> Nollataan muuttuja overk.
         overk = 0;
      else if (block == l && block != blockCount && Score[l] <= k) {
         // Score <= k ==> jatketaan seuraavaan lohkoon. Koska seuraavaa lohkoa
         // ei ole laskettu edellisell sarakkeella, pit sen arvot pivitt.
         Score[l+1] = k+1;
         Pv[l+1] = 0;
         Mv[l+1] = 0;
         overk = 0;
         block = l+1;
      }
      else if (block == l)
         // Seuraavaan lohkoon ei kannata siirty, kun Score > k.
         break;
    }
    blockAverage += block+1;
    if (block == blockCount && Score[blockCount] <= k) {
         if (print)
            printf("Match at %d\n",j);
         count++;
    }
   }
   return count;
   /*ShowMessage("Esiintymi lytyi " + IntToStr(laskuri) + "kpl. " +
               "Aikaa meni " + IntToStr(msec) + "msec");
   ShowMessage("Keskimrinen lohkojen mr " +
           FormatFloat("0.00",(float)blockAverage / n));*/
}

void MCompDistance()
{
   unsigned int Mv = 0;
   unsigned int Pv = (~(unsigned)0) >> (w - m);
   unsigned int Score = m;
   unsigned int Eq, Xv, Xh, Ph, Mh;
   unsigned int twopowmminusone = (unsigned)1 << (m-1);
   int j;
   int i;
   for(j = 1; j <= n-1; j++) {
      Eq = Peq[0][(unsigned char)text[j-1]];
      Xv = Eq | Mv;
      Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
      Ph = Mv | ~ (Xh | Pv);
      Mh = Pv & Xh;
      if (Ph & twopowmminusone)
         Score += 1;
      else if (Mh & twopowmminusone)
         Score -= 1;
      Ph <<= 1;
      Ph |= 1;
      Pv = (Mh << 1) | ~ (Xv | Ph);
      Mv = Ph & Xv;
      if (Score <= k && print)
            printf("Match at %d\n",j);      
      if (j-1+(Score-k) > m + k)
         break;
   }
}


//---------------------------------------------------------------------------

// Navarro

unsigned int shiftr(unsigned int pos,unsigned int tobeshifted,
                    unsigned int bitstoshift)
{
   unsigned int kplusonebitson = (~(unsigned)0) >> (w - (k+1));
   return (kplusonebitson & (tobeshifted >> bitstoshift))
           << (w - pos -(k+1));
}

void NPrecomp(char *P) {
   unsigned int b[256];
   int i;
   unsigned int j;
   for(i = 0;i< 256;i++) {
      S[i] = false;
      b[i] = 0;
      for (j = 1; j <= m; j++)
         if (i == P[j-1]) {
            b[i] |= (1 << (j-1));
            if (j <= k+1)
               S[i] = true;
         }
      b[i] = ~b[i];
      B[i] = 0;
      for (j = 0; j <= m-k-1; j++)
         B[i] = B[i] | shiftr(1+j*(k+2),b[i],j);
      B[i] >>= w - (k+2)*(m-k);
   }
   Din = 0;
   M1 = 0;
   // Alussa (Din) kaikkien diagonaalien arvot ovat 1:st
   // esim. k=2 ==> Din = 01110111011101110111011101110111,
   // koska diagonaalinen pituus voi olla k+1 ja 0 on erottajana.
   // M1:seen talletetaan jokaisen diagonaalin ensimmiseksi arvoksi 1
   // esim. k=2 ==> M1 = 00010001000100010001000100010001.
   // M2:seen viimeinen diagonaali ykkst
   // esim. k=2 ==> M2 = 00010001000100010001000100010111.
   for (j = 0; j <= m-k-1; j++) {
      Din = Din | shiftr(1+j*(k+2),~(unsigned)0,w-(k+1));
      M1 = M1 | shiftr(1+j*(k+2),1,0);
   }
   Din >>= w - (k+2)*(m-k);
   // M2:n voidaan pivitt  M1:n perusteella, eli listn loppuun ykksi
   M2 = M1 | shiftr(1+(m-k-1)*(k+2),~(~(unsigned)0 << (k+1)),0);
   M1 >>= w - (k+2)*(m-k);
   M2 >>= w - (k+2)*(m-k);
   // M3:sta kytetn tyhjentmn viimeinen diagonaali,
   // kun lydetn tsmys.
   // esim. k=2 ==> M3 = 00000000000000000000000000000111.
   M3 = ~(unsigned)0 >> ((m-k-1)*(k+2)+1);
   M3 >>= w - (k+2)*(m-k);
   G = 1 << k;
}

//---------------------------------------------------------------------------

int NComp()
{
   unsigned int x;
   int i=0;
   int count=0;
   D = Din;
   while (++i < n)
      if (S[text[i-1]]) {
         do {
            x = (D >> (k+2)) | B[text[i-1]];
            D = ((D << 1) | M1) & ((D << (k+3)) | M2) &
                (((x+M1) ^ x) >> 1) & Din;
            if ((D & G) == (unsigned)0) {
               if (print)
                  printf("Match at %d\n",i);
               count++;
               D = D | M3;
            }
         } while (D != Din && ++i < n);
      }
   return count;
}

void NCompDist()
{
   int i=0;
   unsigned int x;
   D = Din;
   unsigned zerodiag=0;
   zerolast = 1 << ((k+2)*(m-k)-2);
   zerofirst = 1 << (k+2)*(m-k-1);
   ++i;
     do {
        x = ((D >> (k+2)) | zerodiag) | B[text[i-1]];
        D = ((D << 1) | M1) & ((D << (k+3)) | M2) &
            (((x+M1) ^ x) >> 1) & Din;
        if ((D & G) == (unsigned)0) {
           if (print);
             printf("Match at %d\n",i);
           D = D | M3;
        }
        if ((zerodiag & zerolast) == (unsigned)0)
           zerodiag |= zerodiag + zerofirst;
     } while (D != Din && ++i < n); // jos D=Din yksikn tila ei ole aktiivinen.
}

//---------------------------------------------------------------------------

// SuffixTree

TSTreeNode* g(TSTreeNode* r, char t)
{
   if (r == NULL)
      return NULL;
   if (r->Small())
      return NULL;
   else
      r = ((TSTreeInNode *)r)->firstChild;
   while (r != NULL) {
      if (text[r->GetMinindex()-1] == t)
         return r;
      r = r->branchBroather;
   }
   return NULL;
}

TSTreeNode *gg(TSTreeNode *s,char t, unsigned *kk, unsigned *pp)
{
   if (s == NULL)
      return NULL;
   TSTreeNode* ss;
   if (s->Small())
      ss = NULL;
   else
      ss = ((TSTreeInNode*)s)->firstChild;
   *kk = 0;
   *pp = 0;
   while (ss != NULL) {
      if (ss->GetMinindex() == 0 || text[ss->GetMinindex()-1] == t) {
         *kk = ss->GetMinindex();
         if (ss->GetMinindex() == 0)
            *pp = 0;
         else
            if (ss->Small())
               *pp = ((TSTreeLeaf*)ss)->Maxindex();
            else
               *pp = ((TSTreeInNode*)ss)->Maxindex();
         return ss;
      }
      ss = ss->branchBroather;
   }
   return NULL;
}

void setgg(TSTreeNode *s,unsigned k, unsigned p,TSTreeNode *newnode)
{
  if (((TSTreeInNode*)s)->firstChild == NULL) {
     ((TSTreeInNode*)s)->firstChild = newnode;
     newnode->branchBroather = NULL;
  }
  else {
     newnode->branchBroather = ((TSTreeInNode*)s)->firstChild->branchBroather;
     ((TSTreeInNode*)s)->firstChild->branchBroather = newnode;
  }
  newnode->SetMinindex(k);
}

void setggFirstChild(TSTreeNode *s,unsigned k, unsigned p,TSTreeNode *newnode)
{
  newnode->branchBroather = ((TSTreeInNode*)s)->firstChild;
  ((TSTreeInNode*)s)->firstChild = newnode;
  newnode->SetMinindex(k);
}

void cleargg(TSTreeNode *s,char t)
{
   TSTreeNode *ss;
   ss = ((TSTreeInNode*)s)->firstChild;
   if (ss == NULL)
      return;
   if (text[ss->GetMinindex()-1] == t) {
      ((TSTreeInNode*)s)->firstChild = ss->branchBroather;
   }
   else
      while (ss->branchBroather != NULL) {
         if (text[ss->branchBroather->GetMinindex()-1] == t) {
            ss->branchBroather = ss->branchBroather->branchBroather;
            break;
         }
         ss = ss->branchBroather;
      }
}

bool test_and_split(TSTreeNode *s,unsigned k, unsigned p, char t,
                    TSTreeNode **newnode)
{
   unsigned kk, pp;
   TSTreeNode *ss;
   *newnode = s;
   if (k <= p) {
      ss = gg(s,text[k-1],&kk,&pp);
      if (ss == NULL || kk == 0)
         return false;
      if (t != text[kk+p-k+1-1]) {
         (*newnode) = (TSTreeNode*)(new TSTreeInNode);
         nodeCount++;
         if (((TSTreeInNode*)s)->firstChild == ss) {
            cleargg(s,text[k-1]);
            setggFirstChild(s,kk,kk+p-k,*newnode);
         }
         else {
            cleargg(s,text[k-1]);
            setgg(s,kk,kk+p-k,*newnode);
         }
         setgg(*newnode,kk+p-k+1,pp,ss);
         return false;
      }
      else
         return true;
   }
   else {
      ss = gg(s,t,&kk,&pp);
      if (ss == NULL)
         return false;
      else
         return true;
   }
}

unsigned canonize(TSTreeNode *s,unsigned k, unsigned p,TSTreeNode **r)
{
   TSTreeNode *ss;
   unsigned kk,pp;
   (*r) = s;
   if (p<k) return k;
   else {
      ss = gg(*r,text[k-1],&kk,&pp);
      if (ss == NULL)
         return k;
      while ((int)pp - (int)kk <= (int)p - (int)k) {
         k = k+pp-kk+1;
         *r =ss;
         if (k <= p)
            ss = gg(*r,text[k-1],&kk,&pp);
      }
      return k;
   }
}

unsigned update(TSTreeNode **s,unsigned k, unsigned i)
{
   TSTreeNode *oldr, *r, *rr;
   oldr = root;
   bool endpoint = test_and_split(*s,k,i-1,text[i-1],&r);
   while (!endpoint) {
      rr = (TSTreeNode *)(new TSTreeLeaf);
      nodeCount++;
      if (i > n)
         setgg(r,n,n,rr);
      else
         setgg(r,i,n,rr);
      if (oldr != root)
         ((TSTreeInNode*)oldr)->suffixLink = r;
      oldr = r;
      if (((TSTreeInNode*)*s)->suffixLink == NULL)
         break;
      k = canonize(((TSTreeInNode*)*s)->suffixLink,k,i-1,s);
      endpoint = test_and_split(*s,k,i-1,text[i-1],&r);
   }
   if (oldr != root)
      ((TSTreeInNode*)oldr)->suffixLink = *s;
   return k;
}

void CalcSuffixTreeNodesAtEachDepth(TSTreeNode *r, unsigned depth)
{
   unsigned maxindex;
   unsigned olddepth=depth;
   if (r->Small())
      maxindex = ((TSTreeLeaf*)r)->Maxindex();
   else
      maxindex = ((TSTreeInNode*)r)->Maxindex();
   int j;
   for (j=r->GetMinindex();j<=maxindex && depth < 1000;j++, depth++)
      Nodes[depth]++;
   if (!r->Small() && depth < 1000)
      CalcSuffixTreeNodesAtEachDepth(((TSTreeInNode*)r)->firstChild, depth);
   if (r->branchBroather != NULL)
      CalcSuffixTreeNodesAtEachDepth(r->branchBroather, olddepth);
}

void MakeSuffixTree()
{
   TSTreeNode *empty,*s;
   int nodeNro = 0;
   if (root != NULL)
      delete root;
   empty = (TSTreeNode *)(new TSTreeInNode);
   root = (TSTreeNode *)(new TSTreeInNode);
   nodeCount++;
   ((TSTreeInNode*)root)->suffixLink = empty;
   int j;
   setgg(empty,0,0,root);
   int i;
   for (i=0; i< 256; i++)
      V[i] = 0;
   i=0;
   int k = 1;
   s = root;
   while (i < n) {
      i++;
      if (text[i-1]!='#')
         V[text[i-1]]++; // merkkien jakauma yls
      k = update(&s,k,i);
      k = canonize(s,k,i,&s);
   }
   update(&s,k,i);
   int c=0;
   for (i = 0; i < 256; i++) 
      if (V[i])
         c++; // aakkosto selville
   // tm partitiontihakua varten
   for (i=0; i<1000; i++)
      Nodes[i] = 0;
   CalcSuffixTreeNodesAtEachDepth(((TSTreeInNode*)root)->firstChild,1);
   if (print) {
      printf("Text size is %dB, c=%d\n",n,c);
      printf("Tree takes %d*%dB + %d*%dB = %dB space (%.4fs B/c)\n"
             ,n,sizeof(TSTreeLeaf),nodeCount-n,sizeof(TSTreeInNode),
              n*sizeof(TSTreeLeaf)+(nodeCount-n)*sizeof(TSTreeInNode),
          ((float)(n*sizeof(TSTreeLeaf)+(nodeCount-n)*sizeof(TSTreeInNode)))/n);
   }
   nodeCount = 0;
}
//---------------------------------------------------------------------------

void ShowMatchesFromTree(TSTreeNode *r,unsigned cnt)
{
  if (r == NULL) {
     MatchList->Add(n - cnt);
  }
  else {
     if (r->Small())
        ShowMatchesFromTree(NULL,cnt+ ((TSTreeLeaf*)r)->Maxindex() - r->GetMinindex()+1);
     else
        ShowMatchesFromTree(((TSTreeInNode*)r)->firstChild,cnt+ ((TSTreeInNode*)r)->Maxindex() - r->GetMinindex()+1);
     if (r->branchBroather != NULL)
        ShowMatchesFromTree(r->branchBroather,cnt);
  }
}

void MScanSuffixTree(TSTreeNode *r, unsigned Mv,unsigned Pv,int Score,
                    unsigned depth)
{
   nodesVisited++;
   unsigned Mvold = Mv;
   unsigned Pvold = Pv;
   int Scoreold = Score;
   unsigned depthold = depth;
   int j;
   unsigned maxindex;
   if (r->Small())
      maxindex = ((TSTreeLeaf*)r)->Maxindex();
   else
      maxindex = ((TSTreeInNode*)r)->Maxindex();
   for (j=r->GetMinindex();j<=maxindex;j++) {
      if (Score != m || S[text[j-1]]) {
         Eq = Peq[0][(unsigned char)text[j-1]];
         Xv = Eq | Mv;
         Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
         Ph = Mv | ~ (Xh | Pv);
         Mh = Pv & Xh;
         if (Ph & twopowmminusone)
           Score += 1;
         else if (Mh & twopowmminusone)
           Score -= 1;
         Ph <<= 1;
         Ph |= 1;
         Pv = (Mh << 1) | ~ (Xv | Ph);
         Mv = Ph & Xv;
         if (Score <= k && text[j-1] != '#')
            if (r->Small())
               ShowMatchesFromTree(NULL,maxindex-j);
            else
               ShowMatchesFromTree(((TSTreeInNode*)r)->firstChild,maxindex-j);
      }
      else {
         sumdepths +=depth;
         sumstops++;
         break;
      }
      depth++;
      if (depth-1+(Score-k) > m + k) {
         sumdepths +=depth-1;
         sumstops++;
         break;
      }
   };
   sumnodedepths += depth - depthold > 1 ? depth - depthold : 1;
   if ((j > maxindex) && (!r->Small()))
      MScanSuffixTree(((TSTreeInNode*)r)->firstChild,Mv,Pv,Score,depth);
   if (r->branchBroather != NULL)
      MScanSuffixTree(r->branchBroather,Mvold,Pvold,Scoreold,depthold);
}


void MScanSuffixTreeUnstricted(TSTreeNode *r, unsigned *Mv,unsigned *Pv,unsigned *Score,
                    unsigned depth, unsigned overk, unsigned block)
{
   unsigned *Pvold = (unsigned *)malloc((blockCount+1)*sizeof(unsigned));
   unsigned *Mvold = (unsigned *)malloc((blockCount+1)*sizeof(unsigned));
   unsigned *Scoreold = (unsigned *)malloc((blockCount+1)*sizeof(unsigned));
   int l;
   for (l=0; l <= blockCount; l++) {
      Mvold[l] = Mv[l];
      Scoreold[l] = Score[l];
      Pvold[l] = Pv[l];
   }
   unsigned overkold = overk;
   unsigned blockold = block;
   unsigned depthold = depth;
   unsigned Phendbit;
   unsigned Mhendbit;
   unsigned temp;
   int j;
   int count;
   unsigned maxindex;
   if (r->Small())
      maxindex = ((TSTreeLeaf*)r)->Maxindex();
   else
      maxindex = ((TSTreeInNode*)r)->Maxindex();
   for (j=r->GetMinindex();j<=maxindex;j++) {
      if (block != 0 || Score[block] != (block+1)*w || S[text[j-1]]) {
         Phendbit = 1;
         Mhendbit = 0;
         for (l=0; l <= blockCount; l++) {
            Eq = Peq[l][(unsigned char)text[j-1]];
            Xv = Eq | Mv[l];
            Xh = ((((Eq | Mhendbit)& Pv[l]) + Pv[l]) ^ Pv[l]) | (Eq | Mhendbit);
            Ph = Mv[l] | ~ (Xh | Pv[l]);
            Mh = Pv[l] & Xh;
            temp = l < blockCount ? twopowwpos : twopowmpos;
            if (Ph & temp)
               Score[l] += 1;
            else if (Mh & temp)
               Score[l] -= 1;
            temp = (Ph >> (w-1));
            Ph <<= 1;
            Ph |= Phendbit;
            Phendbit = temp;
            temp = (Mh >> (w-1));
            Mh <<= 1;
            Mh |= Mhendbit;
            Mhendbit = temp;
            Pv[l] = (Mh) | ~ (Xv | Ph);
            Mv[l] = Ph & Xv;
            if (block == l+1 &&
                ((block == blockCount &&
                 (Score[l] > k+lastBlockLength ||
                 (Score[l] > k && overk != 0 && j - overk >= lastBlockLength))) ||
                (block != blockCount &&
                 (Score[l] > k+w ||
                 (Score[l] > k && overk != 0 && j - overk >= w)))))  {
               // lohkon Score kasvanut niin isoksi, ettei seuraavaa kannata laskea
               overk = 0;
               block = l;
               break;
            }
            else if (block == l+1 && overk == 0 && Score[l] > k)
               // Talletetaan ensimminen diagonaali jolla Score > k. Nin tiedetn
               // jatkossa koska seuraavan lohkon laskeminen on turhaa.
               overk = j;
            else if (block == l+1 && Score[l] <= k)
               // Diagonaalilla Score <= k ==> Nollataan muuttuja overk.
               overk = 0;
            else if (block == l && block != blockCount && Score[l] <= k) {
               // Score <= k ==> jatketaan seuraavaan lohkoon. Koska seuraavaa lohkoa
               // ei ole laskettu edellisell sarakkeella, pit sen arvot pivitt.
               Score[l+1] = k+1;
               Pv[l+1] = 0;
               Mv[l+1] = 0;
               overk = 0;
               block = l+1;
            }
            else if (block == l)
               // Seuraavaan lohkoon ei kannata siirty, kun Score > k.
               break;
         }
         if (block == blockCount && Score[blockCount] <= k) {
            if (r->Small())
               ShowMatchesFromTree(NULL,maxindex-j);
            else
               ShowMatchesFromTree(((TSTreeInNode*)r)->firstChild,maxindex-j);
         }
      }
      else {
         sumdepths +=depth;
         sumstops++;
         break;
      }
      depth++;
      if (depth-1+(Score[block]-k) > m + k) {
         sumdepths +=depth-1;
         sumstops++;
         break;
      }
   }
   sumnodedepths += depth - depthold > 1 ? depth - depthold : 1;
   if ((j > maxindex) && (!r->Small()))
      MScanSuffixTreeUnstricted(((TSTreeInNode*)r)->firstChild,Mv,Pv,Score,depth,overk,block);
   else {
      free(Mv);
      free(Score);
      free(Pv);
   }
   if (r->branchBroather != NULL)
      MScanSuffixTreeUnstricted(r->branchBroather,Mvold,Pvold,Scoreold,depthold,overkold,blockold);
   else {
      free(Mvold);
      free(Scoreold);
      free(Pvold);
   }

}

void NScanSuffixTree(TSTreeNode *r, unsigned D,unsigned zerodiag, unsigned depth)
{
   unsigned Dold = D;
   unsigned depthold = depth;
   unsigned zerodiagold = zerodiag;
   bool zerofull=(zerodiag & zerolast) != (unsigned)0;
   unsigned x;
   int j;
   unsigned maxindex;
   if (r->Small())
      maxindex = ((TSTreeLeaf*)r)->Maxindex();
   else
      maxindex = ((TSTreeInNode*)r)->Maxindex();
   for (j=r->GetMinindex();j<=maxindex;j++) {
      if (D == Din && (!S[text[j-1]] || zerofull))
         break;
      x = ((D >> (k+2)) | zerodiag) | B[text[j-1]];
      D = ((D << 1) | M1) & ((D << (k+3)) | M2) &
          (((x+M1) ^ x) >> 1) & Din;
      if ((D & G) == (unsigned)0) {
         D = D | M3;
         if (r->Small())
            ShowMatchesFromTree(NULL,maxindex-j);
         else
            ShowMatchesFromTree(((TSTreeInNode*)r)->firstChild,maxindex-j);
      }
      if (!zerofull || ((zerodiag & zerolast) == (unsigned)0))
         zerodiag |= zerodiag + zerofirst;
      else
         zerofull = true;
      depth++;
      if (depth-1 > m + k)
         break;
   };
   if ((j > maxindex) && (!r->Small()))
      NScanSuffixTree(((TSTreeInNode*)r)->firstChild,D,zerodiag,depth);
   if (r->branchBroather != NULL)
      NScanSuffixTree(r->branchBroather,Dold,zerodiagold,depthold);
}

int MSearchSuffixTree()
{
   twopowmminusone = (unsigned)1 << (m-1);
   nodesVisited =0;
   sumdepths = 0;
   sumstops = 0;
   sumnodedepths = 0;
   MatchList->Clear();
   if (m <=w)
      MScanSuffixTree(((TSTreeInNode*)root)->firstChild, 0, (~(unsigned)0) >> (w - m), m, 1);
   else {
      MPreCompUnstricted(pattern);
      MScanSuffixTreeUnstricted(((TSTreeInNode*)root)->firstChild, Mv, Pv, Score, 1, 0, block);
   }
   int j=0;
   if (debug)
      printf("Nodes visited %d, average depth %f, checked 0.0\n",
              sumnodedepths,sumdepths/(float)sumstops);
   while (print && j < MatchList->Top)
       printf("Match at %d\n", MatchList->Data[j++]);
   return MatchList->Top;
}

char * SubString(char * A,unsigned start, unsigned count)
{
   char * Result = (char*)malloc((count+1)*sizeof(char));
   unsigned i;
   for (i=start; i < start+count && A[i-1] != '\0'; i++)
      Result[i-start] = A[i-1];
   Result[i-start] = '\0';
   return Result;
}

void PrintTree(TSTreeNode *s, int depth)
{
   int i;
   unsigned count;
   char * ApuS=(char *)malloc((depth+1)*sizeof(char));
   for (i=0;i< depth;i++)
      ApuS[i] = ' ';
   ApuS[i] = '\0';
   if (s->Small()) {
      count = ((TSTreeLeaf*)s)->Maxindex()-s->GetMinindex();
      if (count+depth > 30)
         count = 30-depth;
      printf("%s%s\n",ApuS,SubString(text,s->GetMinindex()-1,count));
   }
   else {
      count = ((TSTreeInNode*)s)->Maxindex()-s->GetMinindex();
      if (count+depth > 30)
         count = 30-depth;
      printf("%s%s\n",ApuS,SubString(text,s->GetMinindex()-1,count));
   }
   if (!s->Small())
      PrintTree(((TSTreeInNode*)s)->firstChild,depth+1);
   if (s->branchBroather != NULL)
      PrintTree(s->branchBroather,depth);
}

int NSearchSuffixTree()
{
   MatchList->Clear();
   zerolast = 1 << (k+2)*(m-k)-2;
   zerofirst = 1 << (k+2)*(m-k-1);
   NScanSuffixTree(((TSTreeInNode*)root)->firstChild,Din,0,1);
   int j=0;
   while (print && j < MatchList->Top)
      printf("Match at %d\n",MatchList->Data[j++]);
   return MatchList->Top;
}

//---------------------------------------------------------------------------

int DPSearch(char *P)
{
   int i,j;
   int *C;
   C = (int *)malloc((m+1)*sizeof(int));
   for (i=0; i<=m; i++)
      C[i] = i;
   j =1;
   int Cj,d;
   char t,p;
   int firstkplusone=k+1;
   int limit=firstkplusone;
   int count = 0;
   do {
      Cj = C[0];
      C[0] = 0;
      t = *(text++);
      for (i=1; i<=m; i++) {
         d = minOfThree(Cj+(t == P[i-1] ? 0 : 1),C[i-1]+1,C[i]+1);
         Cj = C[i];
         C[i] = d;
         if (firstkplusone == 0 && d > k)
            firstkplusone = i;
         else if (d <= k)
            firstkplusone = 0;
         if (i >= limit || i == m) {
            limit++;
            if (firstkplusone != 0 && firstkplusone < limit)
               limit = firstkplusone;
            firstkplusone = 0;
            if (limit <= m)
               C[limit] = k+1;
            break;
         }
      }
      if (limit >= m && C[m] <= k) {
         if (print)
            printf("Match at %d \n",j);
         count++;
      }
   } while (j++ <= n-1);
   return count;
}
//---------------------------------------------------------------------------

void PatternPartition(int rr)
{
   int i,len;
   if (PP != NULL) {
      for (i=0; i < part_r; i++)
         free(PP[i]);
      free(PP);
   }
   part_r = rr;
   if (part_r <= 1)
      return;
   PP = (PartStruct **)malloc(part_r*sizeof(PartStruct *));
   PP[0] = (PartStruct *)malloc(sizeof(PartStruct));
   PP[0]->a = 1;
   for (i=0;i < part_r; i++) {
      len = i < m % part_r ? m/part_r + 1 : m/part_r;
      PP[i]->b = PP[i]->a + len - 1;
      PP[i]->error = k/part_r;
      if (i == part_r-1)
         break;
      PP[i+1] = (PartStruct *)malloc(sizeof(PartStruct));
      PP[i+1]->a = PP[i]->b + 1;
   }
}
//---------------------------------------------------------------------------

int MCheckMatchesUnstricted(PartStruct *P)
{
   int count=0;
   unsigned Phendbit;
   unsigned Mhendbit;
   unsigned temp;
   int j;
   int l;
   int i;
   unsigned overk;
   for (i=0; i < MatchList->Top; i++) {
      for (l=0; l < blockCount; l++) {
         Mv[l] = 0;
         Score[l] = (l+1)*w;
         Pv[l] = (~(unsigned)0);
      }
      block = (k-1) / w;
      Mv[blockCount] = 0;
      Score[blockCount] = m;
      overk=1;
      Pv[blockCount] = (~(unsigned)0) >> ((blockCount+1)*w - m);
      j = MatchList->Data[i] - P->b + 1 - k;
      if (j < 1)
         j = 1;
      for(; j <= MatchList->Data[i]-P->b+1+m+k; j++) {
         if (j > n-1)
            break;
         Phendbit = 0;
         Mhendbit = 0;
         for (l=0; l <= blockCount; l++) {
            Eq = Peq[l][(unsigned char)text[j-1]];
            Xv = Eq | Mv[l];
            Xh = ((((Eq | Mhendbit)& Pv[l]) + Pv[l]) ^ Pv[l]) | (Eq | Mhendbit);
            Ph = Mv[l] | ~ (Xh | Pv[l]);
            Mh = Pv[l] & Xh;
            temp = l < blockCount ? twopowwpos : twopowmpos;
            if (Ph & temp)
               Score[l] += 1;
            else if (Mh & temp)
               Score[l] -= 1;
            temp = (Ph >> (w-1));
            Ph <<= 1;
            Ph |= Phendbit;
            Phendbit = temp;
            temp = (Mh >> (w-1));
            Mh <<= 1;
            Mh |= Mhendbit;
            Mhendbit = temp;
            Pv[l] = (Mh) | ~ (Xv | Ph);
            Mv[l] = Ph & Xv;
            if (block == l+1 &&
                ((block == blockCount &&
                  (Score[l] > k+lastBlockLength ||
                   (Score[l] > k && overk != 0 && j - overk >= lastBlockLength))) ||
                 (block != blockCount &&
                  (Score[l] > k+w ||
                   (Score[l] > k && overk != 0 && j - overk >= w)))))  {
               // lohkon Score kasvanut niin isoksi, ettei seuraavaa kannata laskea
               overk = 0;
               block = l;
               break;
            }
            else if (block == l+1 && overk == 0 && Score[l] > k)
               // Talletetaan ensimminen diagonaali jolla Score > k. Nin tiedetn
               // jatkossa koska seuraavan lohkon laskeminen on turhaa.
               overk = j;
            else if (block == l+1 && Score[l] <= k)
               // Diagonaalilla Score <= k ==> Nollataan muuttuja overk.
               overk = 0;
            else if (block == l && block != blockCount && Score[l] <= k) {
               // Score <= k ==> jatketaan seuraavaan lohkoon. Koska seuraavaa lohkoa
               // ei ole laskettu edellisell sarakkeella, pit sen arvot pivitt.
               Score[l+1] = k+1;
               Pv[l+1] = 0;
               Mv[l+1] = 0;
               overk = 0;
               block = l+1;
            }
            else if (block == l)
               // Seuraavaan lohkoon ei kannata siirty, kun Score > k.
               break;
         }
         if (block == blockCount && Score[blockCount] <= k) {
            if (MatchingOnes->IndexOf(j) == -1)
               MatchingOnes->Add(j);
            count++;
        }
     }
   }
   return count;
}

int MCheckMatches(PartStruct *P)
{
   unsigned int Mv;
   unsigned int Pv;
   unsigned int Score = m;
   unsigned int Eq, Xv, Xh, Ph, Mh;
   unsigned int twopowmminusone = (unsigned)1 << (m-1);
   int j=0;
   int i;
   int count=0;
   for (i=0; i < MatchList->Top; i++) {
      Mv = 0;
      Pv = (~(unsigned)0) >> (w - m);
      j = MatchList->Data[i] - P->b + 1 - k;
      Score = m;
      if (j < 1)
         j = 1;
      for(; j <= MatchList->Data[i]-P->b+1+m+k; j++) {
         if (j > n-1)
            break;
         Eq = Peq[0][(unsigned char)text[j-1]];
         Xv = Eq | Mv;
         Xh = (((Eq & Pv) + Pv) ^ Pv) | Eq;
         Ph = Mv | ~ (Xh | Pv);
         Mh = Pv & Xh;
         if (Ph & twopowmminusone)
            Score += 1;
         else if (Mh & twopowmminusone)
            Score -= 1;
         Ph <<= 1;
         Pv = (Mh << 1) | ~ (Xv | Ph);
         Mv = Ph & Xv;
         if (Score <= k) {
            if (MatchingOnes->IndexOf(j) == -1)
               MatchingOnes->Add(j);
            count++;
         }
      }
   }
   return count;
}

int MPartSearchST(char *P)
{
   int i,j;
   unsigned oldm = m;
   unsigned oldk = k;
   MatchingOnes = new TMatchList();
   int count=0;
   nodesVisited =0;
   sumdepths = 0;
   sumstops = 0;
   sumnodedepths = 0;
   int sumchecks = 0;
   for (i=0; i<part_r; i++) {
      m = PP[i]->b-PP[i]->a + 1;
      k = PP[i]->error;
      MatchList->Clear();
      twopowmminusone = (unsigned)1 << (m-1);
      if (m <= w) {
         MPreComp(SubString(pattern,PP[i]->a,m));
         MScanSuffixTree(((TSTreeInNode*)root)->firstChild, 0, (~(unsigned)0) >> (w - m), m, 1);
      }
      else {
         MPreCompUnstricted(SubString(P,PP[i]->a,m));
         MScanSuffixTreeUnstricted(((TSTreeInNode*)root)->firstChild, Mv, Pv, Score, 1,0,block);
      }
      m = oldm;
      k = oldk;
      sumchecks += MatchList->Top;
      if (m <= w) {
         MPreComp(P);
         MCheckMatches(PP[i]);
      }
      else {
         MPreCompUnstricted(P);
         MCheckMatchesUnstricted(PP[i]);
      }
   }
   if (debug)
      printf("Nodes visited %d, average depth %f, checked %d\n",
              sumnodedepths,sumdepths/(float)sumstops,sumchecks);
   count = MatchingOnes->Top;
   if (print)
      for (i = 0; i < MatchingOnes->Top; i++)
         printf("Match at %d\n",MatchingOnes->Data[i]);
   delete MatchingOnes;
   return count;
}



//---------------------------------------------------------------------------

int CompareSuffixes(char *S, unsigned x, unsigned y, unsigned depth)
{
   /*if (depth != 0) {
      if (x+depth-1 < n && y+depth-1 < n && S[x+depth-1] < S[y+depth-1])
         return -1;
      else if (x+depth-1 < n && y+depth-1 < n && S[x+depth-1] > S[y+depth-1])
         return +1;
      else if (x+depth-1 < n && y+depth-1 < n)
         return 0;
      else
         return x < y ? -1: +1;
   }
   else {*/
      int inc=0;
      while (x+inc+depth < n && y+inc+depth < n && S[x + inc+depth] == S[y + inc+depth])
         inc++;
      if (x+inc+depth == n)
         return +(inc+1);
      else if (y+inc == n)
         return -(inc+1);
      else
         return S[x+inc+depth] < S[y+inc+depth] ? -(inc+1) : +inc+1;
   //}
}

void QuickSort(TSAElement *A, unsigned First, unsigned Last, unsigned depth)
{
  unsigned First_old = First;
  unsigned Last_old = Last;
  unsigned posMiddle = A[(First_old + Last_old) / 2].pos;
  do {
    while (A[First].pos != posMiddle && CompareSuffixes(text,A[First].pos, posMiddle, 0) < 0)
       First++;
    while (A[Last].pos != posMiddle && Last >= First_old && CompareSuffixes(text,A[Last].pos, posMiddle, 0) > 0) {
       if (Last == 0)
          break;
       Last--;
    }
    if (First <= Last) {
      TSAElement temp = A[First];
      A[First] = A[Last];
      A[Last] = temp;
      First++;
      if (Last == 0)
         break;
      Last--;
    }
  } while (First <= Last);
  if (Last > First_old)
     QuickSort(A, First_old, Last, 0);
  if (First < Last_old)
     QuickSort(A, First, Last_old, 0);
}

/*void QuickSort(unsigned *A, unsigned First, unsigned Last, unsigned depth)
{
  unsigned First_old = First;
  unsigned Last_old = Last;
  unsigned posMiddle = A[(First_old + Last_old) / 2];
  unsigned mindepthleft=256;
  unsigned mindepthright=256;
  int templ=257;
  int tempr=257;
  do {
    if (A[First] != posMiddle) {
       templ = CompareSuffixes(text,A[First], posMiddle, depth);
       if (templ < 0)
          mindepthleft = min(mindepthleft,abs(templ)-1);
    }
    while (A[First] != posMiddle &&  templ < 0) {
       First++;
       if (A[First] == posMiddle)
          break;
       templ = CompareSuffixes(text,A[First], posMiddle, depth);
       if (templ < 0)
          mindepthleft = min(mindepthleft,abs(templ)-1);
    }
    if (A[Last] != posMiddle && Last >= First_old) {
       tempr = CompareSuffixes(text,A[Last], posMiddle, depth);
       if (tempr > 0)
          mindepthright = min(mindepthright,abs(tempr)-1);
    }
    while (A[Last] != posMiddle && Last >= First_old && tempr > 0) {
       if (Last == 0)
          break;
       Last--;
       if (A[Last] == posMiddle || Last < First_old)
          break;
       tempr = CompareSuffixes(text,A[Last], posMiddle, depth);
       if (tempr > 0)
          mindepthright = min(mindepthright,abs(tempr)-1);

    }
    if (First <= Last) {
      unsigned temp = A[First];
      A[First] = A[Last];
      A[Last] = temp;
      mindepthleft = min(mindepthleft,abs(tempr)-1);
      mindepthright = min(mindepthright,abs(templ)-1);
      First++;
      if (Last == 0)
         break;
      Last--;
    }
  } while (First <= Last);
  if (Last > First_old)
     QuickSort(A, First_old, Last, (mindepthleft > 255 ? depth: depth+mindepthleft));
  if (First < Last_old)
     QuickSort(A, First, Last_old, (mindepthright > 255 ? depth: depth+mindepthright));
}*/

void QuickSort(unsigned *A, unsigned First, unsigned Last, unsigned depth)
{
  unsigned First_old = First;
  unsigned Last_old = Last;
  unsigned posMiddle = A[(First_old + Last_old) / 2];
  do {
    while (A[First] != posMiddle &&  CompareSuffixes(text,A[First], posMiddle, depth) < 0)
       First++;
    while (A[Last] != posMiddle && Last >= First_old && CompareSuffixes(text,A[Last], posMiddle, depth) > 0) {
       if (Last == 0)
          break;
       Last--;
    }
    if (First <= Last) {
      unsigned temp = A[First];
      A[First] = A[Last];
      A[Last] = temp;
      First++;
      if (Last == 0)
         break;
      Last--;
    }
  } while (First <= Last);
  if (Last > First_old)
     QuickSort(A, First_old, Last, 0);
  if (First < Last_old)
     QuickSort(A, First, Last_old, 0);
}



unsigned remap(int f,int l,int k,int j) {
    return f <= j - k ? j : l + f + k - j;
}

void SwitchAreas(TSAElement *A,unsigned start, unsigned end, unsigned where)
{
   unsigned i=0;
   TSAElement *temp=(TSAElement *)malloc((end-start+1)*sizeof(TSAElement));
   for (i=0; i <= end-start; i++)
      temp[i] = A[start+i];
   for (i=start-1; i > where; i--)
      A[i+(end-start)+1] = A[i];
   A[where + (end-start)+1] = A[where];
   for (i=0; i <= end-start; i++)
      A[where+i] = temp[i];
   free(temp);
}

void ReOrderPart(TSAElement *A,unsigned start, unsigned end, unsigned depth)
{
   if (start+1 >= end)
      return;
   unsigned i;
   i = start;
   while (A[i].depth < depth)
      i++;
   unsigned newstart = i;
   while (i + A[i].jump < end+1) {
      i = i+ A[i].jump;
      if (depth >= A[i].depth) {
         ReOrderPart(A,newstart,i-1, depth+1);
         newstart = i;
      }
   }
   if (depth == 0)
      return;
   i = start;
   newstart = start;
   unsigned mindepthpos;
   unsigned mindepth = A[i+1].depth;
   unsigned newend;
   while (i < end) {
      mindepthpos = i;
      while (i <= end) {
         if (A[i].jump > 1 && i+1 <=end && A[i].depth <= A[i+1].depth && A[i+1].depth < mindepth) {
            mindepthpos = i;
            mindepth = A[i+1].depth;
         }
         else if (A[i].jump == 1 && A[i].depth < mindepth) {
            mindepthpos = i;
            mindepth = A[i].depth;
         }
         i++;
      }
      if (newstart != mindepthpos) {
         unsigned temp = A[mindepthpos].depth;
         A[mindepthpos].depth = A[newstart].depth;
         A[newstart].depth = temp;
         i = 1;
         while (i+mindepthpos <= end && A[i+mindepthpos].depth > depth)
            i++;
         SwitchAreas(A,mindepthpos,i+mindepthpos-1,newstart);
         newstart = newstart +i;
         mindepth = A[newstart+1].depth;
         i = newstart;
      }
      else {
         newstart = newstart + A[newstart].jump;
         mindepth = A[newstart+1].depth;
         i = newstart;
      }
   }
}

void ReOrderSuffixArray()
{
   unsigned i,j,k,oldk;
   unsigned First = 0;
   unsigned Last;
   char depth;
   TSAElement temp;
   for (i=0; i<n-1; i++) {
      SuffixArray[i].depth = abs(CompareSuffixes(text,SuffixArray[i].pos,
                             SuffixArray[i+1].pos,0))-1;
      SuffixArray[i].jump = 1;
      if (((int)SuffixArray[i].depth == 0 && i > 0 && First <= i-1) ||
           i == n-2) {
         Last = i+(i==n-2 && SuffixArray[n-2].depth > 0 ? 1:0);
         j = Last-1;
         oldk = 0;
         k = 0;
         while (Last >= First) {
            depth = SuffixArray[remap(First,Last,k,Last)].depth;
            // jos lcp(j-1) >= lcp(j) ja lcp(j) >= lcp(Last)
            // ==> nykyist seuraava aloittaa lohkon
            if (j > First) {
               while (j > First && SuffixArray[remap(First,Last,k,j)].depth >
                      SuffixArray[remap(First,Last,k,j-1)].depth &&
                      SuffixArray[remap(First,Last,k,j)].depth >= depth)
                  j--;
               j++;
            }
            if (Last == First)
               j == Last;
            for (k=0; k+j <= Last; k++) {
               temp = SuffixArray[remap(First,Last,oldk+k,First+oldk+k)];
               SuffixArray[remap(First,Last,oldk+k,First+oldk+k)] =
                  SuffixArray[remap(First,Last,oldk+k,j+k)];
               if (k == 0)
                  if (Last-j+1 > 255)
                     SuffixArray[remap(First,Last,oldk+k,First+oldk)].jump = 255;
                  else
                     SuffixArray[remap(First,Last,oldk+k,First+oldk)].jump=(Last-j+1);
               SuffixArray[remap(First,Last,oldk+k,j+k)] = temp;
            }
            oldk += k;
            if (j <= 1)
               break;
            Last = (--j)--;
         }
         First = i+1;
      }
      else
         SuffixArray[i].jump = 1;
   }
   SuffixArray[n-1].depth = 0;
   SuffixArray[n-1].jump = 1;
}

void UpdateSuffixArray()
{
   unsigned i;
   SuffixArray[0].depth = 0;
   // pivitetn syvyys
   for (i=1; i<n; i++)
      SuffixArray[i].depth = abs(CompareSuffixes(text,SuffixArray[i-1].pos,
                             SuffixArray[i].pos,0))-1;
   // pivitetn hyppylinkit pisimpn mahdolliseen hyppyyn
   unsigned maxjump;
   for (i=n-2; i>=0 && i <= n-2; i--)
      if (SuffixArray[i].jump > 0) {
         maxjump = 1;
         if (SuffixArray[i].depth > SuffixArray[i+1].depth) {
            SuffixArray[i].jump = maxjump;
            continue;
         }
         while (SuffixArray[i+maxjump].depth >= SuffixArray[i+1].depth &&
                SuffixArray[i+maxjump].jump > 0) {
            maxjump += SuffixArray[i+maxjump].jump;
         }
         SuffixArray[i].jump = maxjump;
      }
}

void CreateSuffixArray()
{
   SuffixArray = (TSAElement *)malloc(n*(sizeof(TSAElement)));
   unsigned i;
   for (i=0; i<n; i++) {
      SuffixArray[i].pos = i;
      SuffixArray[i].jump = 1;
      SuffixArray[i].depth = 0;
   }
   QuickSort(SuffixArray,0,n-1,0);
}

void MakeSuffixArray()
{
   CreateSuffixArray();
   ReOrderSuffixArray();
   UpdateSuffixArray();
   //ReOrderPart(SuffixArray,0,n-1,0);
   //UpdateSuffixArray();
   if (print) {
      printf("Array takes %d*%d B = %d B space \n",
             n,sizeof(unsigned)+ 2*sizeof(unsigned char),
             n*sizeof(unsigned)+2*sizeof(unsigned char));
      printf("Text size is %d B\n",n);
   }
}

void CalcDepthCactus()
{
   if (DepthCactus != NULL)
      free(DepthCactus);
   DepthCactus = (unsigned char *)malloc(n*sizeof(unsigned char));
   unsigned i;
   DepthCactus[0] = 0;
   for (i=1; i<n; i++)
      DepthCactus[i] = min(255,abs(CompareSuffixes(text,SuffixCactus[i-1],
                             SuffixCactus[i],0))-1);
}

void CalcSiblingCactus()
{
   if (SiblingCactus != NULL)
      free(SiblingCactus);
   SiblingCactus = (unsigned *)malloc(n*sizeof(unsigned));
   SiblingCactus[0] = 1;
   unsigned s;
   unsigned skminusone=0;
   unsigned siplusone;
   unsigned si;
   unsigned r;
   for (s=1; s <= n-1; s++) {
      if (DepthCactus[s-1] <= DepthCactus[s+1-1]) {
         SiblingCactus[s+1-1] = skminusone;
         skminusone = s;
      }
      else {
         siplusone = s;
         si = skminusone;
         while (DepthCactus[si-1] > DepthCactus[s+1-1]) {
            r = SiblingCactus[si+1-1];
            SiblingCactus[si+1-1] = siplusone;
            siplusone = si;
            si = r;
         }
         SiblingCactus[s+1-1] = siplusone;
         skminusone = si;
      }
   }
   siplusone = n;
   si = skminusone;
   while (si > 0) {
      r = SiblingCactus[si+1-1];
      SiblingCactus[si+1-1] = siplusone;
      siplusone = si;
      si = r;
   }
}

void CreateSuffixCactus()
{
   if (SuffixCactus != NULL)
      free(SuffixCactus);
   SuffixCactus = (unsigned *)malloc(n*sizeof(unsigned));
   unsigned i;
   for (i=0; i<256; i++)
      V[i]=0;
   for (i=0; i<n; i++) {
      SuffixCactus[i] = i;
      if (text[i] != '#')
         V[text[i]]++;
   }
   QuickSort(SuffixCactus,0,n-1,0);
}

void MakeSuffixCactus()
{
   CreateSuffixCactus();
   CalcDepthCactus();
   CalcSiblingCactus();
   int c=0;
   int i;
   for (i = 0; i < 256; i++)
      if (V[i])
         c++;
   if (print) {
      printf("Text size is %d B, c=%d\n",n,c);
      printf("Array takes %d*%d B = %d B space (%.4fs B/c)\n",
             n,2*sizeof(unsigned)+ sizeof(unsigned char),
             n*(2*sizeof(unsigned)+ sizeof(unsigned char)),
             (n*(2*sizeof(unsigned)+ sizeof(unsigned char)))/(float)n);
   }
}

//---------------------------------------------------------------------------

void PrintSuffixArray()
{
   unsigned i;
   for (i=0; i<255 && i <n; i++);
      printf("%s (%d) (%d) \n",SubString(text,SuffixArray[i].pos,20),
             (int)SuffixArray[i].depth,SuffixArray[i].jump);
}
//---------------------------------------------------------------------------

void ShowMatchesFromArray(unsigned j, unsigned depth)
{
   // cout << "Match at " << SuffixArray[j++].pos+depth << "\n";
   MatchList->Add(SuffixArray[j++].pos+depth);
   while (j < n && SuffixArray[j].depth >= depth) {
      if (SuffixArray[j].pos+depth < n)
         MatchList->Add(SuffixArray[j].pos+depth);
         //cout << "Match at " << SuffixArray[j].pos+depth << "\n";
      j++;
   }
}


void MScanSuffixArray()
{
   unsigned depth=0;
   unsigned i;
   unsigned j=0;
   TMSAJump *jump=(TMSAJump *)malloc((m+k+2)*sizeof(TMSAJump));
   jump[0].Score = m;
   jump[0].Pv = (~(unsigned)0) >> (w - m);
   jump[0].Mv = 0;
   unsigned Start=SuffixArray[0].pos;
   while (j < n) {
      for (i=Start; i<n; i++) {
         if (jump[depth].Score != m || S[text[i]]) {
            depth++;
            Eq = Peq[0][(unsigned char)text[i]];
            Xv = Eq | jump[depth-1].Mv;
            Xh = (((Eq & jump[depth-1].Pv) + jump[depth-1].Pv) ^ jump[depth-1].Pv) | Eq;
            Ph = jump[depth-1].Mv | ~ (Xh | jump[depth-1].Pv);
            Mh = jump[depth-1].Pv & Xh;
            if (Ph & twopowmminusone)
              jump[depth].Score = jump[depth-1].Score + 1;
            else if (Mh & twopowmminusone)
              jump[depth].Score = jump[depth-1].Score - 1;
            else
              jump[depth].Score = jump[depth-1].Score;
            Ph <<= 1;
            Ph |= 1;
            jump[depth].Pv = (Mh << 1) | ~ (Xv | Ph);
            jump[depth].Mv = Ph & Xv;
            if (jump[depth].Score <= k && text[i-1] != '#')
               ShowMatchesFromArray(j,depth);
         }
         else
            break;
         if (depth+(jump[depth].Score-k) > m + k +1)
            break;
      }
      if (j+1 >= n)
         break;
      if (SuffixArray[j+1].depth >= depth) {
         // ohitetaan turhat hypyt nopeasti
         j = j + SuffixArray[j].jump;
         while (j < n && SuffixArray[j].depth > depth)
            j = j + SuffixArray[j].jump;
         if (j >=n)
            break;
         if (depth > SuffixArray[j].depth)
            depth = SuffixArray[j].depth;
         Start = SuffixArray[j].pos+depth;
      }
      else {
         j++;
         depth = SuffixArray[j].depth;
         Start = SuffixArray[j].pos+depth;
      }
   }
}

int MSearchSuffixArray()
{
   twopowmminusone = (unsigned)1 << (m-1);
   MatchList->Clear();
   MScanSuffixArray();
   int j=0;
   while (print && j < MatchList->Top)
      printf("Match at %d\n",MatchList->Data[j++]);
   return MatchList->Top;
}

void ShowMatchesFromCactus(unsigned j, unsigned depth)
{
   MatchList->Add(SuffixCactus[j++] + depth);
   while (j < n && DepthCactus[j] >= depth) {
      if (SuffixCactus[j]+depth < n)
         MatchList->Add(SuffixCactus[j]+depth);
         //cout << "Match at " << SuffixArray[j].pos+depth << "\n";
      j++;
   }
}

void MScanSuffixCactus()
{
   unsigned i;
   unsigned depth = 0;
   TMCactusState *MCactusState=(TMCactusState *)malloc((m+k+2)*sizeof(TMCactusState));
   MCactusState[0].Score = m;
   MCactusState[0].Pv = (~(unsigned)0) >> (w - m);
   MCactusState[0].Mv = 0;
   TCactusStack *CactusStack = new TCactusStack();
   unsigned j=0;
   while (j < n) {
      depth = DepthCactus[j];
      for (i=SuffixCactus[j]+depth; i<n; i++) {
         depth++;
         if (MCactusState[depth-1].Score != m || S[text[i]]) {
            Eq = Peq[0][(unsigned char)text[i]];
            Xv = Eq | MCactusState[depth-1].Mv;
            Xh = (((Eq & MCactusState[depth-1].Pv) + MCactusState[depth-1].Pv) ^ MCactusState[depth-1].Pv) | Eq;
            Ph = MCactusState[depth-1].Mv | ~ (Xh | MCactusState[depth-1].Pv);
            Mh = MCactusState[depth-1].Pv & Xh;
            if (Ph & twopowmminusone)
              MCactusState[depth].Score = MCactusState[depth-1].Score + 1;
            else if (Mh & twopowmminusone)
              MCactusState[depth].Score = MCactusState[depth-1].Score - 1;
            else
              MCactusState[depth].Score = MCactusState[depth-1].Score;
            Ph <<= 1;
            Ph |= 1;
            MCactusState[depth].Pv = (Mh << 1) | ~ (Xv | Ph);
            MCactusState[depth].Mv = Ph & Xv;
            if (MCactusState[depth].Score <= k && text[i] != '#')
               ShowMatchesFromCactus(j,depth);
         }
         else
            break;
         if (depth+(MCactusState[depth].Score-k) > m + k)
            break;
      }
      i = j+1;
      int depthprev=-1;
      while (i < n && SiblingCactus[i]-1 > j && DepthCactus[SiblingCactus[i]-1] < depth
             && DepthCactus[SiblingCactus[i]-1] > depthprev) {
         i = SiblingCactus[i]-1;
         depthprev = DepthCactus[i];
         CactusStack->Push(i);
      }
      if (CactusStack->Empty())
         break;
      j = CactusStack->Pop();
   }
   free(MCactusState);
   delete CactusStack;
}

void MScanSuffixCactusUnstricted()
{
   unsigned i;
   unsigned depth = 0;
   unsigned Phendbit;
   unsigned Mhendbit;
   int block = (k-1) / w;
   unsigned overk = 0;
   TMCactusStateUnstricted *MCactusState=(TMCactusStateUnstricted *)malloc((m+k+2)*sizeof(TMCactusStateUnstricted));
   for (i=0; i < m+k+2; i++) {
      MCactusState[i].Pv = (unsigned *)malloc((blockCount+1)*sizeof(unsigned));
      MCactusState[i].Mv = (unsigned *)malloc((blockCount+1)*sizeof(unsigned));
      MCactusState[i].Score = (unsigned *)malloc((blockCount+1)*sizeof(unsigned));
      MCactusState[i].block = blockCount;
   }
   for (i=0; i<=blockCount; i++) {
      MCactusState[0].Score[i] = Score[i];
      MCactusState[0].Pv[i] = Pv[i];
      MCactusState[0].Mv[i] = Mv[i];
   }
   MCactusState[0].block = block;
   TCactusStack *CactusStack = new TCactusStack();
   unsigned temp;
   int l;
   unsigned j=0;
   while (j < n) {
      depth = DepthCactus[j];
      block = MCactusState[depth].block;
      for (i=SuffixCactus[j]+depth; i<n; i++) {
         depth++;
         MCactusState[depth].block = MCactusState[depth-1].block;
         if (block != 0 || MCactusState[depth-1].Score[block] !=(unsigned)(block+1)*w || S[text[i]]) {
            Phendbit = 1;
            Mhendbit = 0;
            for (l=0; l <= blockCount; l++) {
               Eq = Peq[l][(unsigned char)text[i]];
               Xv = Eq | MCactusState[depth-1].Mv[l];
               Xh = ((((Eq | Mhendbit)& MCactusState[depth-1].Pv[l]) + MCactusState[depth-1].Pv[l]) ^ MCactusState[depth-1].Pv[l]) | (Eq | Mhendbit);
               Ph = MCactusState[depth-1].Mv[l] | ~ (Xh | MCactusState[depth-1].Pv[l]);
               Mh = MCactusState[depth-1].Pv[l] & Xh;
               temp = l < blockCount ? twopowwpos : twopowmpos;
               if (Ph & temp)
                  MCactusState[depth].Score[l] = MCactusState[depth-1].Score[l]+1;
               else if (Mh & temp)
                  MCactusState[depth].Score[l] = MCactusState[depth-1].Score[l]-1;
               else
                  MCactusState[depth].Score[l] = MCactusState[depth-1].Score[l];
               temp = (Ph >> (w-1));
               Ph <<= 1;
               Ph |= Phendbit;
               Phendbit = temp;
               temp = (Mh >> (w-1));
               Mh <<= 1;
               Mh |= Mhendbit;
               Mhendbit = temp;
               MCactusState[depth].Pv[l] = (Mh) | ~ (Xv | Ph);
               MCactusState[depth].Mv[l] = Ph & Xv;
               if (block == l+1 &&
                   ((block == blockCount &&
                    (MCactusState[depth].Score[l] > k+lastBlockLength ||
                    (MCactusState[depth].Score[l] > k && overk != 0 && j - overk >= lastBlockLength))) ||
                   (block != blockCount &&
                    (MCactusState[depth].Score[l] > k+w ||
                    (MCactusState[depth].Score[l] > k && overk != 0 && j - overk >= w)))))  {
                  // lohkon Score kasvanut niin isoksi, ettei seuraavaa kannata laskea
                  overk = 0;
                  block = l;
                  MCactusState[depth].block = l;
                  break;
               }
               else if (block == l+1 && overk == 0 && MCactusState[depth].Score[l] > k)
                  // Talletetaan ensimminen diagonaali jolla Score > k. Nin tiedetn
                  // jatkossa koska seuraavan lohkon laskeminen on turhaa.
                  overk = j;
               else if (block == l+1 && MCactusState[depth].Score[l] <= k)
                  // Diagonaalilla Score <= k ==> Nollataan muuttuja overk.
                  overk = 0;
               else if (block == l && block != blockCount && MCactusState[depth].Score[l] <= k) {
                  // Score <= k ==> jatketaan seuraavaan lohkoon. Koska seuraavaa lohkoa
                  // ei ole laskettu edellisell sarakkeella, pit sen arvot pivitt.
                  MCactusState[depth-1].Score[l+1] = k+1;
                  MCactusState[depth-1].Pv[l+1] = 0;
                  MCactusState[depth-1].Mv[l+1] = 0;
                  overk = 0;
                  MCactusState[depth].block = l+1;
                  block = l+1;
               }
               else if (block == l)
                  // Seuraavaan lohkoon ei kannata siirty, kun Score > k.
                  break;
            }
            if (block == blockCount && MCactusState[depth].Score[blockCount] <= k)
               ShowMatchesFromCactus(j,depth);
         }
         else
            break;
         if (depth+(MCactusState[depth].Score[block]-k) > m + k)
            break;
      }
      i = j+1;
      int depthprev=-1;
      while (i <= n && SiblingCactus[i]-1 > j && DepthCactus[SiblingCactus[i]-1] < depth
             && DepthCactus[SiblingCactus[i]-1] > depthprev) {
         i = SiblingCactus[i]-1;
         depthprev = DepthCactus[i];
         CactusStack->Push(i);
      }
      if (CactusStack->Empty())
         break;
      j = CactusStack->Pop();
   }
   delete CactusStack;
   for (i=1; i < m+k+2; i++) {
      free(MCactusState[i].Pv);
      free(MCactusState[i].Mv);
      free(MCactusState[i].Score);
   }
   free(MCactusState);
}

void NScanSuffixCactus()
{
   unsigned i;
   unsigned depth = 0;
   TNCactusState *NCactusState=(TNCactusState *)malloc((m+k+2)*sizeof(TNCactusState));
   NCactusState[0].D = Din;
   NCactusState[0].zerodiag = 0;
   TCactusStack *CactusStack = new TCactusStack();
   bool zerofull = false;
   unsigned j=0;
   unsigned x;
   while (j < n) {
      depth = DepthCactus[j];
      for (i=SuffixCactus[j]+depth; i<n; i++) {
         depth++;
         if (NCactusState[depth-1].D == Din && (!S[text[i]] || ((NCactusState[depth-1].zerodiag & zerolast) != (unsigned)0)))
            break;
         x = ((NCactusState[depth-1].D >> (k+2)) | NCactusState[depth-1].zerodiag) | B[text[i]];
         NCactusState[depth].D = ((NCactusState[depth-1].D << 1) | M1) & ((NCactusState[depth-1].D << (k+3)) | M2) &
             (((x+M1) ^ x) >> 1) & Din;
         if ((NCactusState[depth].D & G) == (unsigned)0) {
            NCactusState[depth].D = NCactusState[depth].D | M3;
            ShowMatchesFromCactus(j,depth);
         }
         if ((NCactusState[depth-1].zerodiag & zerolast) == (unsigned)0)
            NCactusState[depth].zerodiag = NCactusState[depth-1].zerodiag | (NCactusState[depth-1].zerodiag + zerofirst);
         else
            NCactusState[depth].zerodiag = NCactusState[depth-1].zerodiag;
         if (depth > m + k)
            break;
      }
      i = j+1;
      int depthprev=-1;
      while (i < n && SiblingCactus[i]-1 > j && DepthCactus[SiblingCactus[i]-1] < depth
             && DepthCactus[SiblingCactus[i]-1] > depthprev) {
         i = SiblingCactus[i]-1;
         depthprev = DepthCactus[i];
         CactusStack->Push(i);
      }
      if (CactusStack->Empty())
         break;
      j = CactusStack->Pop();
   }
   free(NCactusState);
   delete CactusStack;
}

int MSearchSuffixCactus()
{
   twopowmminusone = (unsigned)1 << (m-1);
   MatchList->Clear();
   if (m <=w)
      MScanSuffixCactus();
   else {
      MPreCompUnstricted(pattern);
      MScanSuffixCactusUnstricted();
   }
   int j=0;
   while (print && j < MatchList->Top)
       printf("Match at %d\n", MatchList->Data[j++]);
   return MatchList->Top;
}

int MPartSearchSC(char *P)
{
   int i,j;
   unsigned oldm = m;
   unsigned oldk = k;
   MatchingOnes = new TMatchList();
   int count=0;
   int sumchecks=0;
   for (i=0; i<part_r; i++) {
      m = PP[i]->b-PP[i]->a + 1;
      k = PP[i]->error;
      MatchList->Clear();
      twopowmminusone = (unsigned)1 << (m-1);
      if (m <= w) {
         MPreComp(SubString(pattern,PP[i]->a,m));
         MScanSuffixCactus();
      }
      else {
         MPreCompUnstricted(SubString(P,PP[i]->a,m));
         MScanSuffixCactusUnstricted();
      }
      m = oldm;
      k = oldk;
      sumchecks += MatchList->Top;
      if (m <= w) {
         MPreComp(P);
         MCheckMatches(PP[i]);
      }
      else {
         MPreCompUnstricted(P);
         MCheckMatchesUnstricted(PP[i]);
      }
   }
   count = MatchingOnes->Top;
   if (print)
      for (i = 0; i < MatchingOnes->Top; i++)
         printf("Match at %d\n",MatchingOnes->Data[i]);
   delete MatchingOnes;
   return count;
}

int DPPartSearchSC(char *P)
{
}

int DPPartSearchST(char *P)
{
}

void DPScanSuffixTree(TSTreeNode* r, int *C, int limit, int depth, char *P)
{
   int *Cold=(int*)malloc((m+1)*sizeof(int));
   int limitold = limit;
   int depthold = depth;
   int i;
   for (i=0; i<=m; i++)
      Cold[i]=C[i];
   int j;
   char t;
   int Cj,d;
   int firstkplusone=0;
   unsigned maxindex;
   if (r->Small())
      maxindex = ((TSTreeLeaf*)r)->Maxindex();
   else
      maxindex = ((TSTreeInNode*)r)->Maxindex();
   for (j=r->GetMinindex();j<=maxindex;j++) {
      Cj = C[0];
      C[0] = depth+1;
      t = text[j-1];
      if (depth == 0 && !S[t])
         break;
      for (i=1; i<=m; i++) {
         d = minOfThree(Cj+(t == P[i-1] ? 0 : 1),C[i-1]+1,C[i]+1);
         Cj = C[i];
         C[i] = d;
         if (firstkplusone == 0 && C[i] > k)
            firstkplusone = i;
         else if (C[i] <= k)
            firstkplusone = 0;
         if (i >= limit || i == m) {
            limit++;
            if (firstkplusone != 0 && firstkplusone < limit)
               limit = firstkplusone;
            firstkplusone = 0;
            if (limit <= m)
               C[limit] = k+1;
            break;
         }
      }
      if (limit >= m && C[m] <= k) {
         if (r->Small())
            ShowMatchesFromTree(NULL,maxindex-j);
         else
            ShowMatchesFromTree(((TSTreeInNode*)r)->firstChild,maxindex-j);
      }
      depth++;
      if (depth-1 + (C[min(limit,m)] -k) > m + k)
         break;
   }
   if ((j > maxindex) && (!r->Small()))
      DPScanSuffixTree(((TSTreeInNode*)r)->firstChild,C,limit,depth,P);
   else
      free(C);
   if (r->branchBroather != NULL)
      DPScanSuffixTree(r->branchBroather,Cold,limitold,depthold,P);
   else
      free(Cold);
}

int DPSearchST(char *P)
{
   int *C=(int*)malloc((m+1)*sizeof(int));
   int i,j;
   for (i=0; i<=m; i++)
      C[i] = i;
   for(i = 0;i< 256;i++) {
      S[i] = false;
      for (j = 1; j <= m; j++)
         if (i == P[j-1] && j <= k+1)
            S[i] = true;
   }
   MatchList->Clear();
   DPScanSuffixTree(((TSTreeInNode*)root)->firstChild,C,k+1,0,P);
   if (print)
      for (i=0; i<MatchList->Top; i++)
         printf("Match at %d\n", MatchList->Data[i]);
   return MatchList->Top;
}

void DPScanSuffixCactus(char *P)
{
   unsigned i,j,l;
   unsigned depth = 0;
   TDPCactusState *DPCactusState=(TDPCactusState *)malloc((m+k+2)*sizeof(TDPCactusState));
   for (i=0; i<m+k+2; i++) {
      DPCactusState[i].C = (int *)malloc((m+1)*sizeof(int));
   }
   for (i=0; i<=m; i++)
      DPCactusState[0].C[i] = i;
   TCactusStack *CactusStack = new TCactusStack();
   j=0;
   char t;
   int Cj;
   int firstkplusone=k+1;
   DPCactusState[0].limit = firstkplusone;
   while (j < n) {
      depth = DepthCactus[j];
      for (i=SuffixCactus[j]+depth; i<n; i++) {
         depth++;
         Cj = DPCactusState[depth-1].C[0];
         DPCactusState[depth].C[0] = depth;
         t = text[i];
         if (depth == 1 && !S[t])
            break;
         for (l=1; l<=m; l++) {
            DPCactusState[depth].C[l] = minOfThree(Cj+(t == P[l-1] ? 0 : 1),DPCactusState[depth].C[l-1]+1,DPCactusState[depth-1].C[l]+1);
            Cj = DPCactusState[depth-1].C[l];
            if (firstkplusone == 0 && DPCactusState[depth].C[l] > k)
               firstkplusone = l;
            else if (DPCactusState[depth].C[l] <= k)
               firstkplusone = 0;
            if (l >= DPCactusState[depth-1].limit || l == m) {
               DPCactusState[depth].limit = DPCactusState[depth-1].limit+1;
               if (firstkplusone != 0 && firstkplusone < DPCactusState[depth].limit)
                  DPCactusState[depth].limit = firstkplusone;
               firstkplusone = 0;
               if (DPCactusState[depth].limit <= m)
                  DPCactusState[depth].C[DPCactusState[depth].limit] = k+1;
               break;
            }
         }
         if (DPCactusState[depth].limit >= m && DPCactusState[depth].C[m] <= k) {
            ShowMatchesFromCactus(j,depth);
         }
         if (depth + (DPCactusState[depth].C[min(DPCactusState[depth].limit,m)] -k) > m + k)
            break;
      }
      i = j+1;
      int depthprev=-1;
      while (i < n && SiblingCactus[i]-1 > j && DepthCactus[SiblingCactus[i]-1] < depth
             && DepthCactus[SiblingCactus[i]-1] > depthprev) {
         i = SiblingCactus[i]-1;
         depthprev = DepthCactus[i];
         CactusStack->Push(i);
      }
      if (CactusStack->Empty())
         break;
      j = CactusStack->Pop();
   }
   for (i=0; i<m+k+2; i++)
      free(DPCactusState[i].C);
   free(DPCactusState);
   delete CactusStack;
}

int DPSearchSC(char *P)
{
   int i,j;
   for(i = 0;i< 256;i++) {
      S[i] = false;
      for (j = 1; j <= m; j++)
         if (i == P[j-1] && j <= k+1)
            S[i] = true;
   }
   MatchList->Clear();
   DPScanSuffixCactus(P);
   if (print)
      for (i=0; i<MatchList->Top; i++)
         printf("Match at %d\n", MatchList->Data[i]);
   return MatchList->Top;
}

int NSearchSuffixCactus()
{
   zerolast = 1 << (k+2)*(m-k)-2;
   zerofirst = 1 << (k+2)*(m-k-1);
   MatchList->Clear();
   NScanSuffixCactus();
   int i;
   if (print)
      for (i=0; i<MatchList->Top; i++)
         printf("Match at %d\n", MatchList->Data[i]);
   return MatchList->Top;
}

int NPartSearchSC(char *P)
{
   int i,j;
   unsigned oldm = m;
   unsigned oldk = k;
   MatchingOnes = new TMatchList();
   int count=0;
   int sumchecks=0;
   for (i=0; i<part_r; i++) {
      m = PP[i]->b-PP[i]->a + 1;
      k = PP[i]->error;
      if ((k+2)*(m-k) > w)
         continue;
      MatchList->Clear();
      NPrecomp(SubString(pattern,PP[i]->a,m));
      zerolast = 1 << (k+2)*(m-k)-2;
      zerofirst = 1 << (k+2)*(m-k-1);
      NScanSuffixCactus();
      m = oldm;
      k = oldk;
      sumchecks += MatchList->Top;
      if (m <= w) {
         MPreComp(P);
         MCheckMatches(PP[i]);
      }
      else {
         MPreCompUnstricted(P);
         MCheckMatchesUnstricted(PP[i]);
      }
   }
   count = MatchingOnes->Top;
   if (print)
      for (i = 0; i < MatchingOnes->Top; i++)
         printf("Match at %d\n",MatchingOnes->Data[i]);
   delete MatchingOnes;
   return count;
}


//---------------------------------------------------------------------------


int NPartSearchST(char *P)
{
   int i;
   MatchList->Clear();
   unsigned oldm = m;
   unsigned oldk = k;
   int count=0;
   for (i=0; i<part_r; i++) {
      m = PP[i]->b-PP[i]->a + 1;
      k = PP[i]->error;
      if ((k+2)*(m-k) > w)
         continue;
      zerolast = 1 << (k+2)*(m-k)-2;
      zerofirst = 1 << (k+2)*(m-k-1);
      NPrecomp(SubString(P,PP[i]->a,m));
      MatchList->Clear();
      NScanSuffixTree(((TSTreeInNode*)root)->firstChild,Din,0,1);
      m = oldm;
      k = oldk;
      if (m <= w) {
         MPreComp(P);
         MCheckMatches(PP[i]);
      }
      else {
         MPreCompUnstricted(P);
         MCheckMatchesUnstricted(PP[i]);
      }
   }
   count = MatchingOnes->Top;
   if (print)
      for (i = 0; i < MatchingOnes->Top; i++)
         printf("Match at %d\n",MatchingOnes->Data[i]);
   delete MatchingOnes;
   return count;
}


char * IntToBits(unsigned x) {
   unsigned int i;
   unsigned int yks=1;
   char *Result = (char *)malloc(33*sizeof(char));
   for (i=w;i>0;i--)
      if ((x & (yks << (i-1))) != 0)
         Result[i-1] = '1';
      else
         Result[i] = '0';
   Result[w] = '\0';
   return Result;
}

long double factorial(int start,int stop)
{
   int i;
   long double Result=1.0;
   for (i=start; i<=stop; i++)
      Result *=i;
   return Result;
}

long double AoverB(int a, int b)
{
   return factorial(max(b,a-b)+1,a)/factorial(1,min(b,a-b));
}

long double powld(long double a, long double b)
{
   int i;
   long double Result=1.0;
   for (i=1; i <= b; i++)
      Result *= a;
   return Result;
}

float sigma(int mm, int kk, int c) 
{
   float errorrate = kk/(float)mm;
   float temp = 1-errorrate;
   float x;
   if (kk == 0)
      x = 1/(float)c;
   else
      x = 1/(c*pow((double)errorrate,(double)2*errorrate/temp)*temp*temp);
   return pow((double)x,(double)temp);
}

float CalcProbOfMatchN(int mmm, int kkk, int c)
{
   return pow((double)sigma(mmm,kkk,c),(double)mmm);   
}

float CalcProbOfMatchExact(int mmm,int s,int kkk,int c)
{
   int i,j,p,r;
   float Result=0.0;
   float ***M=(float***)malloc(sizeof(float**)*(mmm+1)); 
   for (i=0; i <= mmm; i++) {
      M[i] = (float**)malloc(sizeof(float*)*(mmm+kkk+1));
      for (j=0; j <= mmm+kkk; j++)
         M[i][j] = (float*)malloc(sizeof(float)*(kkk+1));
   }
   for (i=0; i <= mmm; i++) 
      for (p=0; p <= kkk; p++) 
         if (i <= p) 
            M[i][0][p] = 1.0;
         else
            M[i][0][p] = 0.0;
   for (j=0; j <= mmm+kkk; j++)
      for (p=0; p <= kkk; p++) 
         if (j <= p) 
            M[0][j][p] = 1.0;
         else
            M[0][j][p] = 0.0;
   for (i=0; i <= mmm; i++)
      for (j=0; j <= mmm+kkk; j++)
         if (i == j)    
            M[i][j][0] = 1/(float)pow((double)c,(double)j); 
         else
            M[i][j][0] = 0.0;
   for (i=1; i <= mmm; i++) 
      for (j=1; j <= mmm+kkk; j++)  
         for (p=1; p <= kkk; p++) {
            M[i][j][p] = M[i][j-1][p-1] + M[i-1][j-1][p-1];
            for (r=0; r <= p && r <= i-1; r++)
               M[i][j][p] = M[i][j][p] + M[i-1-r][j-1][p-r] / c;
         }
   if (s >= 0) // tarkka todennkisyys s:n pituiselle jonolle
      Result = M[mmm][s][kkk];
   else { // todennkisyys rajoittamatta esiintymn pituutta
      for (j=max(0,mmm-kkk); j <= mmm+kkk; j++)
         if (M[mmm][j][kkk] > Result)
            Result += M[mmm][j][kkk];
   }
   for (i=0; i <= mmm; i++) {
      for (j=0; j <= mmm+kkk; j++)
         free(M[i][j]);
      free(M[i]);
   }   
   free(M);
   return Result;
}

float CalcEstimatedNodeCount(int mmm,int kkk,int c)
{
   int i,j,p,r;
   float Result = 0.0;
   float ***M=(float***)malloc(sizeof(float**)*(mmm+1)); 
   for (i=0; i <= mmm; i++) {
      M[i] = (float**)malloc(sizeof(float*)*(mmm+kkk+1));
      for (j=0; j <= mmm+kkk; j++)
         M[i][j] = (float*)malloc(sizeof(float)*(mmm+1));
   }       
   for (i=0; i <= mmm; i++) {
      for (p=0; p <= mmm; p++) 
         if (i <= p) 
            M[i][0][p] = 1.0;
         else
            M[i][0][p] = 0.0;
   } 
   for (j=0; j <= mmm+kkk; j++) 
      for (p=0; p <= mmm; p++) 
         if (j <= p) 
            M[0][j][p] = 1.0;
         else
            M[0][j][p] = 0.0;
   for (i=0; i <= mmm; i++)
      for (j=0; j <= mmm+kkk; j++) {
         if (i == j)    
            M[i][j][0] = 1/(float)pow((double)c,(double)j); 
         else
            M[i][j][0] = 0.0;
      } 
   for (i=1; i <=mmm; i++) {
      for (j=1; j <= mmm+kkk; j++)  
         for (p=1; p <= mmm; p++) {
            M[i][j][p] = M[i][j-1][p-1] + M[i-1][j-1][p-1];
            for (r=0; r <= p && r <= i-1; r++) 
               M[i][j][p] = M[i][j][p] + M[i-1-r][j-1][p-r] / c;
            M[i][j][p] = M[i][j][p] > 1.0 ? 1.0 : M[i][j][p]; 
         }
   }
   float sumofall=0.0;
   float sumofprob=0.0;
   float sumNodes=0.0;
   for (j=1; j <= 2*kkk-1; j++) 
      sumNodes += Nodes[j+1]; 
   for (j=2*kkk; j <= mmm+kkk; j++) {
      sumNodes += Nodes[j+1 > 999 ? 999:j+1];
      sumofall += M[mmm][j][mmm+2*kkk-j]*sumNodes;
      sumofprob += M[mmm][j][mmm+2*kkk-j];
   }
   Result = sumofall/sumofprob;
   for (i=0; i <= mmm; i++) {
      for (j=0; j <= mmm+kkk; j++)
         free(M[i][j]);
      free(M[i]);
   }   
   free(M);
   return Result;   
}

float ProbOfPartPattern(unsigned nn, int num, int index, char * pat, int *V)
{
   int i;
   float Result = 0.0;
   for (i=0; i <= PP[index]->b-PP[index]->a; i++)
      Result += (float)V[pat[PP[index]->a+i-1]]/(float)nn;
   Result /= PP[index]->b-PP[index]->a+1;
   //printf("pattern=%s, prob. of %d matches in %s is %f\n",
   //       pat,num,SubString(pat,PP[index]->a,PP[index]->b-PP[index]->a+1),powld(Result,num));
   return powld(Result,num);
}

float CalcProbOfMatchStatistic(unsigned nn, int mmm, int kkk, int c,
      int index, char *pat, int *V)
{
   long double firstHalf=0.0;
   long double secondHalf=0.0;
   unsigned s;
   //for (s=mmm-kkk; s<=mmm; s++)
   //   firstHalf += AoverB(s-2,mmm-kkk-2);
   firstHalf = ProbOfPartPattern(nn,mmm-kkk,index,pat,V)*AoverB(mmm,mmm-kkk);
   for (s=mmm+1; s<=mmm+kkk; s++)
      secondHalf += ProbOfPartPattern(nn,s-kkk,index,pat,V)*AoverB(mmm,s-kkk);
   return firstHalf + secondHalf;
}

float CalcPartitionTimeN(unsigned nn,unsigned mm,unsigned kk,unsigned rr, unsigned c)
{
   float nodesAtDepthk=0.0;
   float checkCount;
   float probMatch;
   float probDepth = (kk/rr+1) / (1.0-1.09/sqrt(c)); //0.9*((int)(2*(kk/rr)+1)/(1-1.09/sqrt(c)));
   float sigmar = sigma(mm/rr, kk/rr, c);
   int i;
   float alln = log10(nn)/log10(c);
   if (mm/rr+kk/rr < alln) {
      for (i=1; i < ceil(probDepth); i++)
         nodesAtDepthk += pow((double)c,(double)i);
      for (i=(int)ceil(probDepth); i <= mm/rr+kk/rr; i++)
         nodesAtDepthk += pow((double)sigmar,(double)i)*pow((double)c,(double)i);
   }
   else if (probDepth > alln) {
      nodesAtDepthk = nn*(probDepth-alln > 1.0 ? probDepth-alln :1.0);
   }
   else {
      for (i=1; i < ceil(probDepth); i++)
         nodesAtDepthk += pow((double)c,(double)i);
      for (i=(int)ceil(probDepth); i <= alln-1; i++)
         nodesAtDepthk += pow((double)sigmar,(double)i)*pow((double)c,(double)i);      
      for (i=(int)floor(alln); i <= mm/rr+kk/rr; i++)
         nodesAtDepthk += pow((double)sigmar,(double)i)*nn;      
   }
   probMatch = CalcProbOfMatchN(mm/rr,kk/rr,c);
   checkCount = (mm+2*kk)*((int)kk/w+1)*nn*probMatch;
   if (debug)
      printf("Nodes %f, depth %f, checks %f\n",rr*nodesAtDepthk,probDepth,rr*probMatch*nn);
   return rr*(nodesAtDepthk+checkCount);
}

float CalcPartitionTimeExact(unsigned nn,unsigned mm,unsigned kk,unsigned rr, unsigned c)
{
   float nodesAtDepthk=0.0;
   float checkCount=0.0;
   float probMatch;
   m = mm;
   k = kk;
   PatternPartition(rr);
   int i;
   if (rr == 1) {
      nodesAtDepthk = CalcEstimatedNodeCount(mm,kk,c);
      checkCount = 0.0;
   }
   else { 
      for (i=0; i < rr; i++) {
         nodesAtDepthk += CalcEstimatedNodeCount(PP[i]->b-PP[i]->a+1,PP[i]->error,c);
         probMatch = CalcProbOfMatchExact(PP[i]->b-PP[i]->a+1,-1,PP[i]->error,c);
         checkCount += nn*probMatch; 
      } 
   }
   if (debug)
      printf("Nodes %f, checks %f\n",nodesAtDepthk,checkCount);
   return nodesAtDepthk+(mm+2*kk)*((int)kk/w+1)*checkCount;
}

float CalcPartitionTimeStatistic(unsigned nn,unsigned mm,unsigned kk,unsigned rr,
      unsigned c, char * pat, int *V)
{
   float nodesAtDepthk=0.0;
   float checkCount;
   float probMatch=0.0;
   float probDepth = (kk/rr+1) / (1.0-1.09/sqrt(c)); 
   float sigmar = sigma(mm/rr, kk/rr, c);
   int i;
   float alln = log10(nn)/log10(c);
   if (mm/rr+kk/rr < alln) {
      for (i=1; i < ceil(probDepth); i++)
         nodesAtDepthk += pow((double)c,(double)i);
      for (i=(int)ceil(probDepth); i <= mm/rr+kk/rr; i++)
         nodesAtDepthk += pow((double)sigmar,(double)i)*pow((double)c,(double)i);
   }
   else if (probDepth > alln) {
      nodesAtDepthk = nn*(probDepth-alln > 1.0 ? probDepth-alln :1.0);
   }
   else {
      for (i=1; i < ceil(probDepth); i++)
         nodesAtDepthk += pow((double)c,(double)i);
      for (i=(int)ceil(probDepth); i <= alln-1; i++)
         nodesAtDepthk += pow((double)sigmar,(double)i)*pow((double)c,(double)i);      
      for (i=(int)floor(alln); i <= mm/rr+kk/rr; i++)
         nodesAtDepthk += pow((double)sigmar,(double)i)*nn;      
   }
   if (rr==1) 
      probMatch = 0.0;
   else {
      for (i=0; i < rr; i++)    
         probMatch += CalcProbOfMatchStatistic(nn,PP[i]->b-PP[i]->a+1,PP[i]->error,c,i,pat,V);
   }
   if (debug)
      printf("Nodes %f, depth %f, checks %f\n",rr*nodesAtDepthk,probDepth,probMatch*nn);
   checkCount = (mm+2*kk)*((int)kk/w+1)*nn*probMatch;
   return rr*nodesAtDepthk+checkCount;
}


int OptPartition(unsigned nn, unsigned mm, unsigned kk, int rr, unsigned c, char *pat, int *V)
{
   int apur;
   int minrr;
   long double mintime=0.0;
   long double rrtime;
   for (apur=1; apur <= kk+1; apur++) {
      if (rr == 0) {
         m = mm;
         k = kk;
         PatternPartition(apur);
         rrtime = CalcPartitionTimeStatistic(nn,mm,kk,apur,c,pat,V);
      }
      else
         rrtime = CalcPartitionTimeExact(nn,mm,kk,apur,c);
      if (debug)
         printf("r=%d ==> takes c*%d time\n",apur,(int)rrtime);
      if (rrtime < mintime || mintime == 0.0) {
         mintime = rrtime;
         minrr = apur;
      }
   }
   return minrr;
}

int OptPartitionStatistic(unsigned nn, unsigned mm, unsigned kk, unsigned c, char *pat, int *V)
{
   int rr;
   int minrr;
   float mintime=0.0;
   float rrtime=0.0;
   for (rr=2; rr <= kk+1; rr++) {
      m = mm;
      k = kk;
      PatternPartition(rr);
      rrtime = CalcPartitionTimeStatistic(nn,mm,kk,rr,c,pat,V);
      if (debug)
         printf("r=%d ==> takes c*%f time\n",rr,rrtime);
      if (rrtime < mintime || mintime == 0.0) {
         mintime = rrtime;
         minrr = rr;
      }
   }
   return minrr;
}


int CalcOptPart(unsigned nn, unsigned mm, unsigned kk, int rr, char *buf)
{
   unsigned c=0;
   int i;
   for (i=0; i < 256; i++)
      if (V[i]) c++;
   n = nn;
   m = mm;
   k = kk;
   pattern = SubString(buf,1,m);
   return OptPartition(nn,mm,kk,rr,c,pattern,V);
}

//---------------------------------------------------------------------------



